"""Gaussian Mixture Model based fitting and Resampling."""

# Author: Thijs van den Berg <thijs@sitmo.com>
# License: MIT

from thermal.resample._interface import _ResampleInterface
from thermal.resample.gmm import ResampleGmm
from thermal.resample.hist import ResampleHist
from thermal.resample.kde import ResampleKde


class Resample(_ResampleInterface):
    """Generic resampling algorithm interface.

    Parameters
    ----------
    amplitude : string, default='gmm'
        The amplitude density resampling method.
    n_components : int, default=3
        The number of mixture components used when amplitude=='gmm.

    Attributes
    ----------
    size_ : int
        The default number of samples to generate.

    Example
    -------
    ::

        import numpy as np
        import thermal as th

        x = np.random.normal(size=10)
        s = th.Resample().fit(x).resample()

        s
        >>> array([ 0.01212549,  0.04772549,  0.08693959, ..., -0.00519905,
           -0.00908192,  0.00756048])

    """

    def __init__(self, amplitude='gmm', n_components=3, *args, **kwargs):
        super(Resample, self).__init__(*args, **kwargs)

        self.amplitude_ = amplitude

        if amplitude == 'gmm':
            self.amp_ = ResampleGmm(n_components)
        elif amplitude == 'kde':
            self.amp_ = ResampleKde()
        elif amplitude == 'hist':
            self.amp_ = ResampleHist()
        else:
            raise ValueError(f'Unknown amplitude_method "{amplitude}".')

    def fit(self, x, **kwargs):
        """Estimate model parameters of the Gaussian Mixtures Resampler.

        Parameters
        ----------
        x : array-like of shape (n_samples)
            List of data points.

        Returns
        -------
        self : object
            The fitted Gaussian Mixtures Resampler.
        """
        self.amp_.fit(x, **kwargs)
        return self

    def resample(self, size=None, **kwargs):
        """Generate random samples.

        Parameters
        ----------
        size : int, default=None
            Number of samples to generate. When omitted the number of samples will be the same as
            the number of samples used to fit.

        Returns
        -------
        X : array, shape (n_samples)
            Randomly generated sample.
        """
        return self.amp_.resample(size=size, **kwargs)



"""Resample interface definition."""
from abc import ABC, abstractmethod


class _ResampleInterface(ABC):
    def __init__(self, *args, **kwargs):
        self.size_ = -1
        self.x_ = None

    @abstractmethod
    def fit(self, x, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def resample(self, size=None, *args, **kwargs):
        raise NotImplementedError

    def _check_fitted(self):
        if self.size_ == -1:
            raise ValueError('This resampler instance is not yet fitted.')



"""Historical sampling based resampling."""

# Author: Thijs van den Berg <thijs@sitmo.com>
# License: MIT

import numpy as np

from thermal.resample._interface import _ResampleInterface


class ResampleHist(_ResampleInterface):
    """Resample using Historical sampling, either with- or without- replacement."""

    def __init__(self, *args, **kwargs):
        super(ResampleHist, self).__init__(*args, **kwargs)
        self.x_ = None

    def fit(self, x, **kwargs):
        """Provide a set of samples that will be used to reample.

        Parameters
        ----------
        x : array-like of shape (n_samples)
            List of data points.

        Returns
        -------
        self : object
            The fitted Historical Resampler.
        """
        self.x_ = x
        self.size_ = len(x)
        return self

    def resample(self, size=None, *, replace=True, **kwargs):
        """Generate random samples.

        Parameters
        ----------
        size : int, default=None
            Number of samples to generate. When omitted the number of samples will be the same as
            the number of samples used to fit.
        replace : boolean, default=True
            Sample with- or without- replacement.

        Returns
        -------
        X : array, shape (n_samples)
            Randomly drawen sample.
        """
        self._check_fitted()
        if size is None:
            size = self.size_
        ans = np.random.choice(self.x_, size=size, replace=replace)
        return ans



"""Pytest module to test stats interfaces."""
import numpy as np
import pytest

from thermal import ResampleHist, anderson_test, ks_test


def test_ks_test_interface():
    """Testing the ks_test interface."""
    x = np.random.normal(size=10)

    s = np.random.normal(size=(10, 4))
    ans = anderson_test(x, s)

    s = [np.random.normal(size=10), np.random.normal(size=10), np.random.normal(size=10)]
    ans = anderson_test(x, s)

    eng = ResampleHist()
    eng.fit(x)
    ans = anderson_test(x, eng)

    return ans


def test_anderson_test_interface():
    """Testing the ks_test interface."""
    x = np.random.normal(size=10)

    s = np.random.normal(size=(10, 4))
    ans = ks_test(x, s)

    s = [np.random.normal(size=10), np.random.normal(size=10), np.random.normal(size=10)]
    ans = ks_test(x, s)

    eng = ResampleHist()
    eng.fit(x)
    ans = ks_test(x, eng)

    with pytest.raises(Exception):
        ans = ks_test(x, "banana")

    return ans



"""Thermal Modules."""

__author__ = """Thijs van den Berg"""
__email__ = 'thijs@sitmo.com'
__version__ = '0.6.1'

from thermal.resample import Resample, ResampleGmm, ResampleHist, ResampleKde
from thermal.stats import anderson_test, ks_test



"""Two sample test to test if two sets of samples are comming from the same distribution."""

import numpy as np
from scipy.stats import anderson_ksamp, ks_2samp

from thermal.resample._interface import _ResampleInterface


def _surrogates_generator(surrogates, num_tests=100):
    if isinstance(surrogates, _ResampleInterface):
        for _ in range(num_tests):
            s = surrogates.resample()
            yield s
    elif isinstance(surrogates, list):
        for s in surrogates:
            yield s
    elif isinstance(surrogates, np.ndarray):
        for i in range(surrogates.shape[1]):
            s = surrogates[:, i]
            yield s
    else:
        raise ValueError('Unknown surrogates argument type.')


def _generic_test(original, surrogates, tester, num_tests=1000):
    p_values = []
    for s in _surrogates_generator(surrogates, num_tests):
        p_values.append(tester(original, s))
    return p_values


"""
def _generic_adversarial_test(original, surrogates, tester, tester2=None, num_tests=100, num_repeats=4):
    inner_p_values = []
    cross_p_values = []
    hn = int(len(original) / 2)
    for s in _surrogates_generator(surrogates, num_tests):
        for r in range(num_repeats):
            np.random.shuffle(original)
            np.random.shuffle(s)
            inner_p_values.append(tester(original[:hn], original[hn : 2 * hn]))
            cross_p_values.append(tester(original[:hn], s[:hn]))
    if tester2 is None:
        tester2 = tester
    return tester2(inner_p_values, cross_p_values), inner_p_values, cross_p_values
"""


def _ks_test_func(a, b):
    t = ks_2samp(a, b)
    return t.pvalue


def ks_test(original, surrogates, *, num_tests=1000):
    """Run multiple two-sample Kolmogorov–Smirnov test.

    This test is used to test the similarity between the distribution of original data samples
    v.s. multiple set of generated surrogate data samples.

    Parameters
    ----------
    original : Orignal set samples
    surrogates : Sets of samples, or a fitted resampler opject that will be used to generate surrogate samples.
    num_tests : (optional) number of tests to do.

    Returns
    -------
    Array of p-values.

    """
    return _generic_test(original, surrogates, _ks_test_func, num_tests=num_tests)


def _anderson_test_sig_func(a, b):
    statistic, critical_values, significance_level = anderson_ksamp([a, b])
    return significance_level


"""
def _anderson_test_stat_func(a, b):
    statistic, critical_values, significance_level = anderson_ksamp([a, b])
    return statistic
"""


def anderson_test(original, surrogates, *, num_tests=1000):
    """Run multiple two-sample Anderson-Darling test between the original and surrogate samples.

    Tests the null hypothesis that multiple surrogate sets and a set of original samples are drawn
    from the same population without having to specify the distribution function of that population.
    For each test it returns an approximate significance level at which the null hypothesis for the
    provided samples can be rejected. The value is floored / capped at 0.1% / 25%.

    Parameters
    ----------
    original : Orignal set samples
    surrogates : Sets of samples, or a fitted resampler opject.
    num_tests : (optional) number of tests to do.

    Returns
    -------
    Array of significance level values.

    """
    return _generic_test(original, surrogates, _anderson_test_sig_func, num_tests)



"""Pytest module to test interfaces."""

import numpy as np
import pytest

from thermal import Resample, ResampleGmm, ResampleHist, ResampleKde


def test_resample_abc():
    """Testing the ABC size_ attribute."""
    obj = ResampleKde()
    return obj.size_


def test_resample_gmm_interface():
    """Testing the ResampleGmm interface."""
    x = np.random.normal(size=10)

    eng = ResampleGmm()
    eng.fit(x)
    y = eng.resample()
    y = eng.resample(5)

    eng = ResampleGmm(3)
    eng.fit(x)
    y = eng.resample()
    y = eng.resample(5)

    eng = ResampleGmm().fit(x)
    y = eng.resample()
    y = eng.resample(5)

    y = ResampleGmm().fit(x).resample()
    y = ResampleGmm(3).fit(x).resample()
    y = ResampleGmm().fit(x).resample(5)
    return y


def test_resample_hist_interface():
    """Testing the ResampleHist inerface."""
    x = np.random.normal(size=10)

    eng = ResampleHist()
    eng.fit(x)
    y = eng.resample(replace=False)
    y = eng.resample(5)
    y = eng.resample(5, replace=False)

    eng = ResampleHist().fit(x)
    y = eng.resample()

    y = ResampleGmm().fit(x).resample()
    return y


def test_resample_kde_interface():
    """Testing the ResampleKde interface."""
    x = np.random.normal(size=10)

    eng = ResampleKde()
    eng.fit(x)
    y = eng.resample()
    y = eng.resample(5)

    eng = ResampleKde().fit(x)
    y = eng.resample()

    y = ResampleKde().fit(x).resample()
    return y


def test_resample_generic_interface():
    """Testing the ResampleKde interface."""
    x = np.random.normal(size=50)

    y = Resample(amplitude='gmm', n_components=5).fit(x).resample(5)
    y = Resample(amplitude='kde').fit(x).resample(5)
    y = Resample(amplitude='hist').fit(x).resample(5, replace=True)

    with pytest.raises(Exception):
        y = Resample(amplitude='banana').fit(x).resample(5, replace=True)

    with pytest.raises(Exception):
        y = Resample(amplitude='gmm').resample(5, replace=True)
    return y



"""The resample module implements resampling algorithms."""

from thermal.resample.generic import Resample
from thermal.resample.gmm import ResampleGmm
from thermal.resample.hist import ResampleHist
from thermal.resample.kde import ResampleKde



"""Gaussian Mixture Model based fitting and Resampling."""

# Author: Thijs van den Berg <thijs@sitmo.com>
# License: MIT

from sklearn.mixture import GaussianMixture

from thermal.resample._interface import _ResampleInterface


class ResampleGmm(_ResampleInterface):
    """Resample using Gaussian Mixtures.

    Parameters
    ----------
    n_components : int, default=3
        The number of mixture components.

    Attributes
    ----------
    size_ : int
        The default number of samples to generate.
    gmm_ : sklearn.mixture.GaussianMixture object.
        The Gaussian Mixture model.

    Example
    -------
    ::

        import numpy as np
        import thermal as th

        x = np.random.normal(size=10)
        s = th.ResampleGmm(3).fit(x).resample()

        s
        >>> array([ 0.01212549,  0.04772549,  0.08693959, ..., -0.00519905,
           -0.00908192,  0.00756048])

    """

    def __init__(self, n_components=3, *args, **kwargs):
        super(ResampleGmm, self).__init__(*args, **kwargs)

        self.gmm_ = None
        self.n_components_ = n_components

    def fit(self, x, **kwargs):
        """Estimate model parameters of the Gaussian Mixtures Resampler.

        Parameters
        ----------
        x : array-like of shape (n_samples)
            List of data points.

        Returns
        -------
        self : object
            The fitted Gaussian Mixtures Resampler.
        """
        self.size_ = len(x)
        self.gmm_ = GaussianMixture(n_components=self.n_components_).fit(x.reshape(-1, 1))
        return self

    def resample(self, size=None, **kwargs):
        """Generate random samples.

        Parameters
        ----------
        size : int, default=None
            Number of samples to generate. When omitted the number of samples will be the same as
            the number of samples used to fit.

        Returns
        -------
        X : array, shape (n_samples)
            Randomly generated sample.
        """
        self._check_fitted()
        if size is None:
            size = self.size_
        samples = self.gmm_.sample(n_samples=size)
        return samples[0].flatten()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tests', 'thermal', 'thermal.resample', 'thermal.stats']

package_data = \
{'': ['*']}

install_requires = \
['click==8.0.4',
 'myst-parser>=0.17.2,<0.18.0',
 'numpy>=1.22.3,<2.0.0',
 'scipy>=1.8.0,<2.0.0',
 'sklearn>=0.0,<0.1']

extras_require = \
{':extra == "docs"': ['Sphinx>=4.5.0,<5.0.0',
                      'sphinx-rtd-theme>=1.0.0,<2.0.0',
                      'nbsphinx>=0.8.8,<0.9.0'],
 ':extra == "test"': ['black>=22.3.0,<23.0.0'],
 'dev': ['tox>=3.20.1,<4.0.0',
         'virtualenv>=20.2.2,<21.0.0',
         'pip>=20.3.1,<21.0.0',
         'twine>=3.3.0,<4.0.0',
         'pre-commit>=2.12.0,<3.0.0',
         'toml>=0.10.2,<0.11.0',
         'bump2version>=1.0.1,<2.0.0'],
 'test': ['isort>=5.8.0,<6.0.0',
          'flake8>=3.9.2,<4.0.0',
          'flake8-docstrings>=1.6.0,<2.0.0',
          'mypy>=0.900,<0.901',
          'pytest>=6.2.4,<7.0.0',
          'pytest-cov>=2.12.0,<3.0.0']}

entry_points = \
{'console_scripts': ['thermal = thermal.cli:main']}

setup_kwargs = {
    'name': 'thermal',
    'version': '0.6.1',
    'description': 'Surrogate times eries generation.',
    'long_description': '# thermal\n\n[![pypi](https://img.shields.io/pypi/v/thermal.svg)](https://pypi.org/project/thermal/)\n[![python](https://img.shields.io/pypi/pyversions/thermal.svg)](https://pypi.org/project/thermal/)\n[![Build Status](https://github.com/sitmo/thermal/actions/workflows/dev.yml/badge.svg)](https://github.com/sitmo/thermal/actions/workflows/dev.yml)\n[![codecov](https://codecov.io/gh/sitmo/thermal/branch/main/graphs/badge.svg)](https://codecov.io/github/sitmo/thermal)\n\nModel Free surrogate time series generation.\n\n\n\n\n\n* Documentation: <https://sitmo.github.io/thermal>\n* GitHub: <https://github.com/sitmo/thermal>\n* PyPI: <https://pypi.org/project/thermal/>\n* Free software: MIT\n\n\n## Features\n\n* TODO\n',
    'author': 'Thijs van den Berg',
    'author_email': 'thijs@sitmo.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/sitmo/thermal',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8.0,<3.11',
}


setup(**setup_kwargs)



"""Kernel Density based fitting and Resampling."""

# Author: Thijs van den Berg <thijs@sitmo.com>
# License: MIT

import numpy as np

from thermal.resample._interface import _ResampleInterface


class ResampleKde(_ResampleInterface):
    """Resample using Kernel Density estimate.

    Attributes
    ----------
    kernel_width_ : number
        The estimated Kernel width.
    """

    def __init__(self, *args, **kwargs):
        super(ResampleKde, self).__init__(*args, **kwargs)
        self.x_ = None
        self.mu_ = 0.0
        self.sigma_ = 1.0
        self.kernel_width_ = 0.0

    def fit(self, x, **kwargs):
        """Estimate model parameters of the Kernel Density Resampler.

        Parameters
        ----------
        x : array-like of shape (n_samples)
            List of data points.

        Returns
        -------
        self : object
            The fitted Kernel Density Resampler.
        """
        self.size_ = len(x)
        self.x_ = x
        self.mu_ = np.mean(x)
        self.sigma_ = np.std(x, ddof=1)
        self.kernel_width_ = 1.06 * self.sigma_ * len(x) ** -0.2
        return self

    def resample(self, size=None, **kwargs):
        """Generate random samples.

        Parameters
        ----------
         size : int, default=None
             Number of samples to generate. When omitted the number of samples will be the same as
             the number of samples used to fit.

        Returns
        -------
         x : array, shape (n_samples)
             Randomly generated sample.
        """
        self._check_fitted()
        if size is None:
            size = self.size_
        ans = np.random.choice(self.x_, size=size, replace=True) + self.kernel_width_ * np.random.normal(size=size)
        # ans = self.mu_ + (ans - self.mu_) * self.sigma_ * (self.sigma_**2 + self.kernel_width_**2) ** -0.5
        return ans
